import { Draft04 } from "json-schema-library";
import { jsonPointerForPosition } from "./utils/jsonPointers";
import { joinWithOr } from "./utils/formatting";
import getSchema from "./utils/schema-lib/getSchema";
import { debug } from "./utils/debug";
import { el } from "./utils/dom";
/**
 * provides a JSON schema enabled tooltip extension for codemirror
 * @group Codemirror Extensions
 */
export function jsonSchemaHover(schema, options) {
    const hover = new JSONHover(schema, options);
    return async function jsonDoHover(view, pos, side) {
        return hover.doHover(view, pos, side);
    };
}
function formatComplexType(schema, type, draft) {
    return `${type}: ${joinWithOr(schema[type].map((s) => {
        var _a;
        return (_a = s.type) !== null && _a !== void 0 ? _a : draft.resolveRef(s).type;
    }))}`;
}
export class JSONHover {
    constructor(schema, opts) {
        this.opts = opts;
        this.schema = new Draft04(schema);
        this.opts = Object.assign({ parser: JSON.parse }, this.opts);
    }
    getDataForCursor(view, pos, side) {
        const pointer = jsonPointerForPosition(view.state, pos, side);
        let data = undefined;
        // TODO: use the AST tree to return the right hand, data so that we don't have to parse the doc
        try {
            data = this.opts.parser(view.state.doc.toString());
        }
        catch (_a) { }
        if (!pointer) {
            return null;
        }
        // if the data is valid, we can infer a type for complex types
        let subSchema = getSchema(this.schema, pointer, data);
        if (subSchema.type === "error" && data !== undefined) {
            // if the data is invalid, we won't get the type - try again without the data
            subSchema = getSchema(this.schema, pointer, undefined);
            if (subSchema.type === "error") {
                return { pointer };
            }
        }
        return { schema: subSchema, pointer };
    }
    formatMessage(texts) {
        const { message, typeInfo } = texts;
        if (message) {
            return el("div", { class: "cm6-json-schema-hover" }, [
                el("div", {
                    class: "cm6-json-schema-hover--description",
                    text: message,
                }),
                el("div", { class: "cm6-json-schema-hover--code-wrapper" }, [
                    el("code", { class: "cm6-json-schema-hover--code", text: typeInfo }),
                ]),
            ]);
        }
        return el("div", { class: "cm6-json-schema-hover" }, [
            el("div", { class: "cm6-json-schema-hover--code-wrapper" }, [
                el("code", { class: "cm6-json-schema-hover--code", text: typeInfo }),
            ]),
        ]);
    }
    getHoverTexts(data, draft) {
        let typeInfo = null;
        let message = null;
        const { schema } = data;
        if (schema.oneOf) {
            typeInfo = formatComplexType(schema, "oneOf", draft);
        }
        if (schema.anyOf) {
            typeInfo = formatComplexType(schema, "anyOf", draft);
        }
        if (schema.allOf) {
            typeInfo = formatComplexType(schema, "allOf", draft);
        }
        if (schema.type) {
            typeInfo = Array.isArray(schema.type)
                ? joinWithOr(schema.type)
                : schema.type;
        }
        if (schema.description) {
            message = schema.description;
        }
        return { message, typeInfo };
    }
    // return hover state for the current json schema property
    async doHover(view, pos, side) {
        var _a, _b, _c, _d;
        const start = pos, end = pos;
        try {
            const cursorData = this.getDataForCursor(view, pos, side);
            // if we don't have a (sub)schema, we can't show anything
            if (!(cursorData === null || cursorData === void 0 ? void 0 : cursorData.schema))
                return null;
            const getHoverTexts = (_b = (_a = this.opts) === null || _a === void 0 ? void 0 : _a.getHoverTexts) !== null && _b !== void 0 ? _b : this.getHoverTexts;
            const hoverTexts = getHoverTexts(cursorData, this.schema);
            // allow users to override the hover
            const formatter = (_d = (_c = this.opts) === null || _c === void 0 ? void 0 : _c.formatHover) !== null && _d !== void 0 ? _d : this.formatMessage;
            const formattedDom = formatter(hoverTexts);
            return {
                pos: start,
                end,
                arrow: true,
                // to mimic similar modes for other editors
                // otherwise, it gets into a z-index battle with completion/etc
                above: true,
                create: (view) => {
                    return {
                        dom: formattedDom,
                    };
                },
            };
        }
        catch (err) {
            debug.log(err);
            return null;
        }
    }
}
