import { Draft04 } from "json-schema-library";
import { joinWithOr } from "./utils/formatting";
import { parseJSONDocumentState } from "./utils/parseJSONDocument";
// return an object path that matches with the json-source-map pointer
const getErrorPath = (error) => {
    var _a, _b, _c;
    // if a pointer is present, return without #
    if (((_a = error === null || error === void 0 ? void 0 : error.data) === null || _a === void 0 ? void 0 : _a.pointer) && ((_b = error === null || error === void 0 ? void 0 : error.data) === null || _b === void 0 ? void 0 : _b.pointer) !== "#") {
        return error.data.pointer.slice(1);
    }
    // return plain data.property if present
    if ((_c = error === null || error === void 0 ? void 0 : error.data) === null || _c === void 0 ? void 0 : _c.property) {
        return `/${error.data.property}`;
    }
    // else, return the empty pointer to represent the whole document
    return "";
};
/**
 * Helper for simpler class instantiaton
 * @group Codemirror Extensions
 */
export function jsonSchemaLinter(schema, options) {
    const validation = new JSONValidation(schema, options);
    return function jsonDoValidation(view) {
        return validation.doValidation(view);
    };
}
export class JSONValidation {
    constructor(schema, options) {
        // rewrite the error message to be more human readable
        this.rewriteError = (error) => {
            var _a, _b, _c, _d, _e, _f;
            if (error.code === "one-of-error") {
                return `Expected one of ${joinWithOr((_a = error === null || error === void 0 ? void 0 : error.data) === null || _a === void 0 ? void 0 : _a.errors, (data) => data.data.expected)}`;
            }
            if (error.code === "type-error") {
                return `Expected \`${((_b = error === null || error === void 0 ? void 0 : error.data) === null || _b === void 0 ? void 0 : _b.expected) && Array.isArray((_c = error === null || error === void 0 ? void 0 : error.data) === null || _c === void 0 ? void 0 : _c.expected)
                    ? joinWithOr((_d = error === null || error === void 0 ? void 0 : error.data) === null || _d === void 0 ? void 0 : _d.expected)
                    : (_e = error === null || error === void 0 ? void 0 : error.data) === null || _e === void 0 ? void 0 : _e.expected}\` but received \`${(_f = error === null || error === void 0 ? void 0 : error.data) === null || _f === void 0 ? void 0 : _f.received}\``;
            }
            const message = error.message.replaceAll("#/", "").replaceAll("/", ".");
            return message;
        };
        this.options = Object.assign({ jsonParser: parseJSONDocumentState }, options);
        // TODO: support other versions of json schema.
        // most standard schemas are draft 4 for some reason, probably
        // backwards compatibility
        //
        // ajv did not support draft 4, so I used json-schema-library
        this.schema = new Draft04(schema);
    }
    get schemaTitle() {
        var _a;
        return (_a = this.schema.getSchema().title) !== null && _a !== void 0 ? _a : "json-schema";
    }
    // validate using view as the linter extension signature requires
    doValidation(view) {
        if (!this.schema)
            return [];
        const text = view.state.doc.toString();
        // ignore blank json strings
        if (!text || text.trim().length < 3)
            return [];
        const json = this.options.jsonParser(view.state);
        let errors = [];
        try {
            errors = this.schema.validate(json.data);
        }
        catch (_a) { }
        if (!errors.length)
            return [];
        // reduce() because we want to filter out errors that don't have a pointer
        return errors.reduce((acc, error) => {
            const errorPath = getErrorPath(error);
            const pointer = json.pointers.get(errorPath);
            if (pointer) {
                // if the error is a property error, use the key position
                const isPropertyError = error.name === "NoAdditionalPropertiesError";
                acc.push({
                    from: isPropertyError ? pointer.keyFrom : pointer.valueFrom,
                    to: isPropertyError ? pointer.keyTo : pointer.valueTo,
                    // TODO: create a domnode and replace `` with <code></code>
                    // renderMessage: () => error.message,
                    message: this.rewriteError(error),
                    severity: "error",
                    source: this.schemaTitle,
                });
            }
            return acc;
        }, []);
    }
}
